import json
import logging
import os
import random
import time
import urllib.parse as urlparse
from typing import Callable, Optional

import requests
from requests.exceptions import HTTPError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_random_exponential,
)

LOGGER = logging.getLogger(__name__)
DEBUG = bool(os.getenv("RPA_DEBUG_API"))


def _needs_retry(exc: BaseException) -> bool:
    # Don't retry on some specific error codes or messages.

    # https://www.restapitutorial.com/httpstatuscodes.html
    # 400 - payload is bad and needs to be changed
    # 401 - missing auth bearer token
    # 403 - auth is in place, but not allowed (insufficient privileges)
    # 409 - payload not good for the affected resource
    no_retry_codes = [400, 401, 403, 409]

    if isinstance(exc, RequestsHTTPError):
        if exc.status_code in no_retry_codes:
            return False

        if exc.status_code == 429:
            # We hit the rate limiter, so sleep extra.
            seconds = random.uniform(1, 3)
            LOGGER.warning("Rate limit hit, sleeping: %fs", seconds)
            time.sleep(seconds)

    return True


def _before_sleep_log():
    log_method = LOGGER.log

    def extensive_log(level, msg, *args, **kwargs):
        log_method(level, msg, *args, **kwargs)
        if DEBUG:
            LOGGER.debug(msg, *args)

    # Monkeypatch inner logging function so it produces an exhaustive log when
    # used under the before-sleep logging utility in `tenacity`.
    LOGGER.log = extensive_log
    return before_sleep_log(LOGGER, logging.DEBUG, exc_info=True)


class RequestsHTTPError(HTTPError):
    """Custom `requests` HTTP error with status code and message."""

    def __init__(
        self, *args, status_code: int = 0, status_message: str = "Error", **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.status_code = status_code
        self.status_message = status_message


class Requests:
    """Wrapper over `requests` 3rd-party with error handling and retrying support."""

    def __init__(self, route_prefix: str, default_headers: Optional[dict] = None):
        self._route_prefix = route_prefix
        self._default_headers = default_headers

    def handle_error(self, response: requests.Response):
        resp_status_code = response.status_code
        log_func = LOGGER.critical if resp_status_code // 100 == 5 else LOGGER.debug
        log_func("API response: %s %r", resp_status_code, response.reason)

        if response.ok:
            return

        fields = {}
        try:
            fields = response.json()
            while not isinstance(fields, dict):
                # For some reason we might still get a string from the deserialized
                # retrieved JSON payload. If a dictionary couldn't be obtained at all,
                # it will end up raising `RequestsHTTPError`.
                fields = json.loads(fields)
        except (json.JSONDecodeError, ValueError, TypeError):
            # No `fields` dictionary can be obtained at all.
            LOGGER.critical("No fields were returned by the server")
            try:
                response.raise_for_status()
            except Exception as exc:
                LOGGER.exception(exc)
                raise RequestsHTTPError(exc, status_code=resp_status_code) from exc

        err_status_code = 0
        status_message = "Error"
        try:
            err_status_code = int(fields.get("status", resp_status_code))
            status_message = fields.get("code") or fields.get("error", {}).get(
                "code", "Error"
            )
            reason = fields.get("message") or fields.get("error", {}).get(
                "message", response.reason
            )
            raise HTTPError(f"{err_status_code} {status_message}: {reason}")
        except Exception as exc:
            LOGGER.exception(exc)
            raise RequestsHTTPError(
                str(fields), status_code=err_status_code, status_message=status_message
            ) from exc

    @retry(
        # Retry until either succeed or trying for the fifth time and still failing.
        # So sleep and retry for 4 times at most.
        stop=stop_after_attempt(5),
        # If the exception is no worth retrying or the number of tries is depleted,
        # then re-raise the last raised exception.
        reraise=True,
        # Decide if the raised exception needs retrying or not.
        retry=retry_if_exception(_needs_retry),
        # Produce debugging logging prior to each time we sleep & re-try.
        before_sleep=_before_sleep_log(),
        # Sleep between the tries with a random float amount of seconds like so:
        # 1. [0, 2]
        # 2. [0, 4]
        # 3. [0, 5]
        # 4. [0, 5]
        wait=wait_random_exponential(multiplier=2, max=5),
    )
    def _request(
        self,
        verb: Callable[..., requests.Response],
        url: str,
        *args,
        _handle_error: Optional[Callable[[requests.Response], None]] = None,
        _sensitive: bool = False,
        headers: Optional[dict] = None,
        **kwargs,
    ) -> requests.Response:
        # Absolute URLs override the prefix, so they are safe to be sent as they'll be
        # the same after joining.
        url = urlparse.urljoin(self._route_prefix, url)
        headers = headers if headers is not None else self._default_headers
        handle_error = _handle_error or self.handle_error

        url_for_log = url
        if _sensitive:
            # Omit query from the URL since might contain sensitive info.
            split = urlparse.urlsplit(url_for_log)
            url_for_log = urlparse.urlunsplit(
                [split.scheme, split.netloc, split.path, "", split.fragment]
            )

        LOGGER.debug("%s %r", verb.__name__.upper(), url_for_log)
        response = verb(url, *args, headers=headers, **kwargs)
        handle_error(response)
        return response

    # CREATE
    def post(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.post, *args, **kwargs)

    # RETRIEVE
    def get(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.get, *args, **kwargs)

    # UPDATE
    def put(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.put, *args, **kwargs)

    # DELETE
    def delete(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.delete, *args, **kwargs)



import functools
import os
import urllib.parse as urlparse
from typing import Any, Optional

# Sentinel value for undefined arguments.
UNDEFINED = object()


def required_env(name: str, default: Any = UNDEFINED) -> str:
    """Load required environment variable.

    Args:
        name: Name of the requested environment variable.
        default: Value to return if no such env var is found, otherwise `KeyError` is
            raised.
    """
    val = os.getenv(name, default)
    if val is UNDEFINED:
        raise KeyError(f"Missing required environment variable: {name}")
    return val


class RequiresEnv:
    """Raises a custom error when the required env var isn't available."""

    def __init__(self, error_message):
        self._exception = RuntimeError(error_message)

    @functools.wraps(required_env)
    def __call__(self, *args, **kwargs):
        try:
            return required_env(*args, **kwargs)
        except Exception as exc:
            raise self._exception from exc


def url_join(*parts: str) -> Optional[str]:
    """Join parts of URL and handle missing/duplicate slashes."""
    if not parts:
        return None

    url = ""
    for part in parts:
        url = urlparse.urljoin(url, part.strip("/") + "/")
    return url



import logging
from functools import lru_cache
from typing import Dict, Optional, TypedDict

from ._requests import Requests
from ._utils import RequiresEnv, url_join

LOGGER = logging.getLogger(__name__)


AssetMeta = Dict[str, str]
Payload = TypedDict(
    "Payload", {"type": str, "content_type": Optional[str], "url": Optional[str]}
)
Asset = TypedDict(
    "Asset",
    {
        "id": str,
        "name": str,
        "payload": Payload,
    },
)


class AssetNotFound(KeyError):
    """Raised when the queried asset couldn't be found."""


@lru_cache(maxsize=1)
def get_assets_client():
    """Creates and returns an API client based on the injected env vars in CR."""
    requires_env = RequiresEnv(
        "Asset Storage feature can be used with Control Room only"
    )
    api_url = requires_env("RC_API_URL_V1")
    workspace_id = requires_env("RC_WORKSPACE_ID")
    api_token = requires_env("RC_API_TOKEN_V1", "")  # production availability
    api_key = requires_env("RC_API_KEY", "")  # for local tests only

    if not any([api_token, api_key]):
        raise RuntimeError("API token or key missing")
    auth_header = f"Bearer {api_token}" if api_token else f"RC-WSKEY {api_key}"
    default_headers = {
        "Authorization": auth_header,
        "Content-Type": "application/json",
    }
    route_prefix = url_join(api_url, "workspaces", workspace_id, "assets")
    assets_client = Requests(route_prefix, default_headers=default_headers)
    return assets_client



import logging
import random
import time
from typing import List, Optional, cast

from ._requests import RequestsHTTPError
from ._storage import Asset, AssetMeta, AssetNotFound
from ._storage import get_assets_client as _get_assets_client

__version__ = "0.1.0"
version_info = [int(x) for x in __version__.split(".")]

LOGGER = logging.getLogger(__name__)


def list_assets() -> List[AssetMeta]:
    """List all the existing assets.

    Returns:
        A list of assets where each asset is a dictionary with fields like 'id' and
        'name'.
    """
    assets_client = _get_assets_client()
    assets = assets_client.get("").json()
    LOGGER.info("Found %d assets.", len(assets))
    return assets


def _retrieve_asset_id(name: str) -> str:
    for asset in list_assets():
        if asset["name"] == name:
            asset_id = asset["id"]
            LOGGER.debug("Found existing asset %r with ID: %s", name, asset_id)
            return asset_id

    LOGGER.warning("No asset with name %r found, assuming ID.", name)
    return name


def _get_asset(name: str, raise_if_missing: bool = True) -> Optional[Asset]:
    assets_client = _get_assets_client()
    asset_id = _retrieve_asset_id(name)
    exception = None

    def _handle_error(resp):
        # NOTE(cmin764): The API will return 404 if the asset can't be found on the
        #  server, therefore don't let this raise, nor retry given this custom handler.
        nonlocal exception
        try:
            assets_client.handle_error(resp)
        except RequestsHTTPError as exc:
            if exc.status_code == 404:
                exception = exc
            else:
                raise

    LOGGER.debug("Retrieving asset %r with resulted ID %r.", name, asset_id)
    response = assets_client.get(asset_id, _handle_error=_handle_error)
    if response.ok:
        return cast(Asset, response.json())

    message = f"asset with name {name!r} and resulted ID {asset_id!r} couldn't be found"
    if raise_if_missing:
        raise AssetNotFound(message) from exception

    LOGGER.warning("%s%s!", message[0].upper(), message[1:])
    return None


def get_asset(name: str) -> str:
    """Get the asset's value by providing its `name`.

    Args:
        name: Name of the asset to fetch.

    Returns:
        The previously set value of this asset. (empty if it was never set)

    Raises:
        AssetNotFound: When the queried asset doesn't exist.
    """
    LOGGER.info("Retrieving asset %r.", name)
    payload = _get_asset(name)["payload"]  # type: ignore
    if payload["type"] == "empty":
        LOGGER.warning("Asset %r has no value set!", name)
        return ""

    assets_client = _get_assets_client()
    url = payload["url"]
    return assets_client.get(url, headers={}).text


def _create_asset(name: str):
    LOGGER.debug("Creating new asset with name %r.", name)
    assets_client = _get_assets_client()
    body = {"name": name}
    return assets_client.post("", json=body).json()


def set_asset(name: str, value: str, wait: bool = True):
    """Sets/Creates an asset named `name` with the provided `value`.

    Args:
        name: Name of the existing or new asset to create. (if missing)
        value: The new value set within the asset.
        wait: Blocks until the new value is reflected within the asset.

    Raises:
        AssetNotFound: When the queried asset doesn't exist.
    """
    existing_asset = _get_asset(name, raise_if_missing=False)
    if existing_asset:
        asset_id = existing_asset["id"]
        LOGGER.debug("Updating already existing asset with ID %r.", asset_id)
    else:
        asset_id = _create_asset(name)["id"]
        LOGGER.debug("Updating newly created asset with ID %r.", asset_id)

    assets_client = _get_assets_client()
    body = {"content_type": "text/plain"}
    upload_data = assets_client.post(f"{asset_id}/upload", json=body).json()
    assets_client.put(upload_data["upload_url"], data=value, headers={})

    if wait:
        LOGGER.info(
            "Waiting for the sent value to be set successfully in the %r asset...", name
        )
        upload_url = f"{asset_id}/uploads/{upload_data['id']}"
        while True:
            upload_data = assets_client.get(upload_url).json()
            status = upload_data["status"]
            if status == "pending":
                sleep_time = round(random.uniform(0, 1), 2)
                LOGGER.debug("Asset upload still pending, sleeping %.2f...", sleep_time)
                time.sleep(sleep_time)
                continue
            elif status == "completed":
                break
            elif status == "failed":
                raise ValueError(
                    f"asset {name!r} upload failed with {upload_data['reason']!r}"
                )
            else:
                raise TypeError(f"asset {name!r} got invalid status {status!r}")

    LOGGER.info("Asset with name %r set successfully.", name)


def delete_asset(name: str):
    """Delete an asset by providing its `name`.

    Args:
        name: Name of the asset to delete.

    Raises:
        AssetNotFound: When the queried asset doesn't exist.
    """
    LOGGER.info("Deleting asset %r.", name)
    assets_client = _get_assets_client()
    assets_client.delete(_get_asset(name)["id"])  # type: ignore


__all__ = [
    "AssetNotFound",
    "list_assets",
    "get_asset",
    "set_asset",
    "delete_asset",
]
